svn commit: r270287 - in head: sbin/ifconfig sys/net

Alexander V. Chernikov melifaro at FreeBSD.org
Thu Aug 21 17:54:43 UTC 2014


Author: melifaro
Date: Thu Aug 21 17:54:42 2014
New Revision: 270287
URL: http://svnweb.freebsd.org/changeset/base/270287

Log:
  * Add new net/sff8436.h containing constants used to access
    QSFP+ data via i2c inteface. These constants has been taken
    from SFF-8436 "QSFP+ 10 Gbs 4X PLUGGABLE TRANSCEIVER" standard
    rev 4.8.
  * Add support for printing QSFP+ information from 40G NICs
    such as Chelsio T5.
  
  This commit does not contain ioctl changes necessary for this
  functionality work, there will be another commit soon.
  
  Example:
  cxl1: flags=8843<UP,BROADCAST,RUNNING,SIMPLEX,MULTICAST> metric 0 mtu 1500
          options=ec07bb<RXCSUM,TXCSUM,VLAN_MTU,VLAN_HWTAGGING,JUMBO_MTU,.....>
          ether 00:07:43:28:ad:08
          nd6 options=29<PERFORMNUD,IFDISABLED,AUTO_LINKLOCAL>
          media: Ethernet 40Gbase-LR4 <full-duplex>
          status: active
          plugged: QSFP+ 40GBASE-LR4 (MPO Parallel Optic)
          vendor: OEM PN: OP-QSFP-40G-LR4 SN: 20140318001 DATE: 2014-03-18
          module temperature: 64.06 C voltage: 3.26 Volts
          lane 1: RX: 0.47 mW (-3.21 dBm) TX: 2.78 mW (4.46 dBm)
          lane 2: RX: 0.20 mW (-6.94 dBm) TX: 2.80 mW (4.47 dBm)
          lane 3: RX: 0.18 mW (-7.38 dBm) TX: 2.79 mW (4.47 dBm)
          lane 4: RX: 0.90 mW (-0.45 dBm) TX: 2.80 mW (4.48 dBm)
  
  Tested on:	Chelsio T5
  Tested on:	Mellanox/Huawei passive/active cables/transceivers.
  MFC after:	2 weeks
  Sponsored by:	Yandex LLC

Added:
  head/sys/net/sff8436.h   (contents, props changed)
Modified:
  head/sbin/ifconfig/sfp.c
  head/sys/net/sff8472.h

Modified: head/sbin/ifconfig/sfp.c
==============================================================================
--- head/sbin/ifconfig/sfp.c	Thu Aug 21 17:36:42 2014	(r270286)
+++ head/sbin/ifconfig/sfp.c	Thu Aug 21 17:54:42 2014	(r270287)
@@ -34,11 +34,13 @@ static const char rcsid[] =
 #include <sys/socket.h>
 
 #include <net/if.h>
+#include <net/sff8436.h>
 #include <net/sff8472.h>
 
 #include <math.h>
 #include <err.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
@@ -53,11 +55,16 @@ typedef int (read_i2c)(struct i2c_info *
 struct i2c_info {
 	int s;
 	int error;
+	int bshift;
+	int qsfp;
+	int do_diag;
 	struct ifreq *ifr;
 	read_i2c *f;
-	uint8_t diag_type;
 	char *textbuf;
 	size_t bufsize;
+	int cfd;
+	int port_id;
+	int chip_id;
 };
 
 struct _nv {
@@ -86,11 +93,12 @@ static struct _nv conn[] = {
 	{ 0x20, "HSSDC II" },
 	{ 0x21, "Copper pigtail" },
 	{ 0x22, "RJ45" },
+	{ 0x23, "No separate connector" }, /* SFF-8436 */
 	{ 0, NULL }
 };
 
 /* SFF-8472 Rev. 11.4 table 3.5: Transceiver codes */
-/* 10G Ethernet compliance codes, byte 3 */
+/* 10G Ethernet/IB compliance codes, byte 3 */
 static struct _nv eth_10g[] = {
 	{ 0x80, "10G Base-ER" },
 	{ 0x40, "10G Base-LRM" },
@@ -165,6 +173,21 @@ static struct _nv fc_speed[] = {
 	{ 0, NULL }
 };
 
+/* SFF-8436 Rev. 4.8 table 33: Specification compliance  */
+
+/* 10/40G Ethernet compliance codes, byte 128 + 3 */
+static struct _nv eth_1040g[] = {
+	{ 0x80, "Reserved" },
+	{ 0x40, "10GBASE-LRM" },
+	{ 0x20, "10GBASE-LR" },
+	{ 0x10, "10GBASE-SR" },
+	{ 0x08, "40GBASE-CR4" },
+	{ 0x04, "40GBASE-SR4" },
+	{ 0x02, "40GBASE-LR4" },
+	{ 0x01, "40G Active Cable" },
+	{ 0, NULL }
+};
+
 const char *
 find_value(struct _nv *x, int value)
 {
@@ -194,18 +217,15 @@ find_zero_bit(struct _nv *x, int value, 
 }
 
 static void
-get_sfp_identifier(struct i2c_info *ii, char *buf, size_t size)
+convert_sff_identifier(char *buf, size_t size, uint8_t value)
 {
 	const char *x;
-	uint8_t data;
-
-	ii->f(ii, SFF_8472_BASE, SFF_8472_ID, 1, (caddr_t)&data);
 
 	x = NULL;
-	if (data <= SFF_8472_ID_LAST)
-		x = sff_8472_id[data];
+	if (value <= SFF_8024_ID_LAST)
+		x = sff_8024_id[value];
 	else {
-		if (data > 0x80)
+		if (value > 0x80)
 			x = "Vendor specific";
 		else
 			x = "Reserved";
@@ -215,17 +235,14 @@ get_sfp_identifier(struct i2c_info *ii, 
 }
 
 static void
-get_sfp_connector(struct i2c_info *ii, char *buf, size_t size)
+convert_sff_connector(char *buf, size_t size, uint8_t value)
 {
 	const char *x;
-	uint8_t data;
 
-	ii->f(ii, SFF_8472_BASE, SFF_8472_CONNECTOR, 1, (caddr_t)&data);
-
-	if ((x = find_value(conn, data)) == NULL) {
-		if (data >= 0x0D && data <= 0x1F)
+	if ((x = find_value(conn, value)) == NULL) {
+		if (value >= 0x0D && value <= 0x1F)
 			x = "Unallocated";
-		else if (data >= 0x23 && data <= 0x7F)
+		else if (value >= 0x24 && value <= 0x7F)
 			x = "Unallocated";
 		else
 			x = "Vendor specific";
@@ -235,6 +252,42 @@ get_sfp_connector(struct i2c_info *ii, c
 }
 
 static void
+get_sfp_identifier(struct i2c_info *ii, char *buf, size_t size)
+{
+	uint8_t data;
+
+	ii->f(ii, SFF_8472_BASE, SFF_8472_ID, 1, (caddr_t)&data);
+	convert_sff_identifier(buf, size, data);
+}
+
+static void
+get_sfp_connector(struct i2c_info *ii, char *buf, size_t size)
+{
+	uint8_t data;
+
+	ii->f(ii, SFF_8472_BASE, SFF_8472_CONNECTOR, 1, (caddr_t)&data);
+	convert_sff_connector(buf, size, data);
+}
+
+static void
+get_qsfp_identifier(struct i2c_info *ii, char *buf, size_t size)
+{
+	uint8_t data;
+
+	ii->f(ii, SFF_8436_BASE, SFF_8436_ID, 1, (caddr_t)&data);
+	convert_sff_identifier(buf, size, data);
+}
+
+static void
+get_qsfp_connector(struct i2c_info *ii, char *buf, size_t size)
+{
+	uint8_t data;
+
+	ii->f(ii, SFF_8436_BASE, SFF_8436_CONNECTOR, 1, (caddr_t)&data);
+	convert_sff_connector(buf, size, data);
+}
+
+static void
 printf_sfp_transceiver_descr(struct i2c_info *ii, char *buf, size_t size)
 {
 	char xbuf[12];
@@ -290,49 +343,72 @@ get_sfp_transceiver_class(struct i2c_inf
 	snprintf(buf, size, "%s", tech_class);
 }
 
+static void
+get_qsfp_transceiver_class(struct i2c_info *ii, char *buf, size_t size)
+{
+	const char *tech_class;
+	uint8_t code;
+
+	/* Check 10/40G Ethernet class only */
+	ii->f(ii, SFF_8436_BASE, SFF_8436_CODE_E1040G, 1, (caddr_t)&code);
+	tech_class = find_zero_bit(eth_1040g, code, 1);
+	if (tech_class == NULL)
+		tech_class = "Unknown";
+
+	snprintf(buf, size, "%s", tech_class);
+}
 
+/*
+ * Print SFF-8472/SFF-8436 string to supplied buffer.
+ * All (vendor-specific) strings are padded right with '0x20'.
+ */
 static void
-get_sfp_vendor_name(struct i2c_info *ii, char *buf, size_t size)
+convert_sff_name(char *buf, size_t size, char *xbuf)
 {
-	char xbuf[17], *p;
+	char *p;
 
-	memset(xbuf, 0, sizeof(xbuf));
-	/* ASCII String, right-padded with 0x20 */
-	ii->f(ii, SFF_8472_BASE, SFF_8472_VENDOR_START, 16, xbuf);
 	for (p = &xbuf[16]; *(p - 1) == 0x20; p--)
 		;
 	*p = '\0';
-
 	snprintf(buf, size, "%s", xbuf);
 }
 
 static void
+convert_sff_date(char *buf, size_t size, char *xbuf)
+{
+
+	snprintf(buf, size, "20%c%c-%c%c-%c%c", xbuf[0], xbuf[1],
+	    xbuf[2], xbuf[3], xbuf[4], xbuf[5]);
+}
+
+static void
+get_sfp_vendor_name(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[17];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8472_BASE, SFF_8472_VENDOR_START, 16, xbuf);
+	convert_sff_name(buf, size, xbuf);
+}
+
+static void
 get_sfp_vendor_pn(struct i2c_info *ii, char *buf, size_t size)
 {
-	char xbuf[17], *p;
+	char xbuf[17];
 
 	memset(xbuf, 0, sizeof(xbuf));
-	/* ASCII String, right-padded with 0x20 */
 	ii->f(ii, SFF_8472_BASE, SFF_8472_PN_START, 16, xbuf);
-	for (p = &xbuf[16]; *(p - 1) == 0x20; p--)
-		;
-	*p = '\0';
-
-	snprintf(buf, size, "%s", xbuf);
+	convert_sff_name(buf, size, xbuf);
 }
 
 static void
 get_sfp_vendor_sn(struct i2c_info *ii, char *buf, size_t size)
 {
-	char xbuf[17], *p;
+	char xbuf[17];
 
 	memset(xbuf, 0, sizeof(xbuf));
-	/* ASCII String, right-padded with 0x20 */
 	ii->f(ii, SFF_8472_BASE, SFF_8472_SN_START, 16, xbuf);
-	for (p = &xbuf[16]; *(p - 1) == 0x20; p--)
-		;
-	*p = '\0';
-	snprintf(buf, size, "%s", xbuf);
+	convert_sff_name(buf, size, xbuf);
 }
 
 static void
@@ -343,8 +419,47 @@ get_sfp_vendor_date(struct i2c_info *ii,
 	memset(xbuf, 0, sizeof(xbuf));
 	/* Date code, see Table 3.8 for description */
 	ii->f(ii, SFF_8472_BASE, SFF_8472_DATE_START, 6, xbuf);
-	snprintf(buf, size, "20%c%c-%c%c-%c%c", xbuf[0], xbuf[1],
-	    xbuf[2], xbuf[3], xbuf[4], xbuf[5]);
+	convert_sff_date(buf, size, xbuf);
+}
+
+static void
+get_qsfp_vendor_name(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[17];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_VENDOR_START, 16, xbuf);
+	convert_sff_name(buf, size, xbuf);
+}
+
+static void
+get_qsfp_vendor_pn(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[17];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_PN_START, 16, xbuf);
+	convert_sff_name(buf, size, xbuf);
+}
+
+static void
+get_qsfp_vendor_sn(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[17];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_SN_START, 16, xbuf);
+	convert_sff_name(buf, size, xbuf);
+}
+
+static void
+get_qsfp_vendor_date(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[6];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_DATE_START, 6, xbuf);
+	convert_sff_date(buf, size, xbuf);
 }
 
 static void
@@ -353,33 +468,54 @@ print_sfp_vendor(struct i2c_info *ii, ch
 	char xbuf[80];
 
 	memset(xbuf, 0, sizeof(xbuf));
-	get_sfp_vendor_name(ii, xbuf, 20);
-	get_sfp_vendor_pn(ii, &xbuf[20], 20);
-	get_sfp_vendor_sn(ii, &xbuf[40], 20);
-	get_sfp_vendor_date(ii, &xbuf[60], 20);
+	if (ii->qsfp != 0) {
+		get_qsfp_vendor_name(ii, xbuf, 20);
+		get_qsfp_vendor_pn(ii, &xbuf[20], 20);
+		get_qsfp_vendor_sn(ii, &xbuf[40], 20);
+		get_qsfp_vendor_date(ii, &xbuf[60], 20);
+	} else {
+		get_sfp_vendor_name(ii, xbuf, 20);
+		get_sfp_vendor_pn(ii, &xbuf[20], 20);
+		get_sfp_vendor_sn(ii, &xbuf[40], 20);
+		get_sfp_vendor_date(ii, &xbuf[60], 20);
+	}
 
 	snprintf(buf, size, "vendor: %s PN: %s SN: %s DATE: %s",
 	    xbuf, &xbuf[20],  &xbuf[40], &xbuf[60]);
 }
 
+/*
+ * Converts internal templerature (SFF-8472, SFF-8436)
+ * 16-bit unsigned value to human-readable representation:
+ * 
+ * Internally measured Module temperature are represented
+ * as a 16-bit signed twos complement value in increments of
+ * 1/256 degrees Celsius, yielding a total range of –128C to +128C
+ * that is considered valid between –40 and +125C.
+ *
+ */
 static void
-get_sfp_temp(struct i2c_info *ii, char *buf, size_t size)
+convert_sff_temp(char *buf, size_t size, char *xbuf)
 {
-	char xbuf[2];
+	double d;
 
-	int8_t major;
-	uint8_t minor;
-	int k;
+	d = (double)(int8_t)xbuf[0];
+	d += (double)(uint8_t)xbuf[1] / 256;
 
-	memset(xbuf, 0, sizeof(xbuf));
-	ii->f(ii, SFF_8472_DIAG, SFF_8472_TEMP, 2, xbuf);
+	snprintf(buf, size, "%.2f C", d);
+}
 
-	/* Convert temperature to string according to table 3.13 */
-	major = (int8_t)xbuf[0];
-	minor = (uint8_t)buf[1];
-	k = minor * 1000 / 256;
+/*
+ * Retrieves supplied voltage (SFF-8472, SFF-8436).
+ * 16-bit usigned value, treated as range 0..+6.55 Volts
+ */
+static void
+convert_sff_voltage(char *buf, size_t size, char *xbuf)
+{
+	double d;
 
-	snprintf(buf, size, "%d.%d C", major, k / 100);
+	d = (double)(((uint8_t)xbuf[0] << 8) | (uint8_t)xbuf[1]);
+	snprintf(buf, size, "%.2f Volts", d / 10000);
 }
 
 /*
@@ -387,7 +523,7 @@ get_sfp_temp(struct i2c_info *ii, char *
  * human representation.
  */
 static void
-convert_power(struct i2c_info *ii, char *xbuf, char *buf, size_t size)
+convert_sff_power(struct i2c_info *ii, char *buf, size_t size, char *xbuf)
 {
 	uint16_t mW;
 	double dbm;
@@ -397,11 +533,55 @@ convert_power(struct i2c_info *ii, char 
 	/* Convert mw to dbm */
 	dbm = 10.0 * log10(1.0 * mW / 10000);
 
+	/*
+	 * Assume internally-calibrated data.
+	 * This is always true for SFF-8346, and explicitly
+	 * checked for SFF-8472.
+	 */
+
 	/* Table 3.9, bit 5 is set, internally calibrated */
-	if ((ii->diag_type & 0x20) != 0) {
-		snprintf(buf, size, "%d.%02d mW (%.2f dBm)",
-	    	    mW / 10000, (mW % 10000) / 100, dbm);
-	}
+	snprintf(buf, size, "%d.%02d mW (%.2f dBm)",
+    	    mW / 10000, (mW % 10000) / 100, dbm);
+}
+
+static void
+get_sfp_temp(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[2];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8472_DIAG, SFF_8472_TEMP, 2, xbuf);
+	convert_sff_temp(buf, size, xbuf);
+}
+
+static void
+get_sfp_voltage(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[2];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8472_DIAG, SFF_8472_VCC, 2, xbuf);
+	convert_sff_voltage(buf, size, xbuf);
+}
+
+static void
+get_qsfp_temp(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[2];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_TEMP, 2, xbuf);
+	convert_sff_temp(buf, size, xbuf);
+}
+
+static void
+get_qsfp_voltage(struct i2c_info *ii, char *buf, size_t size)
+{
+	char xbuf[2];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_VCC, 2, xbuf);
+	convert_sff_voltage(buf, size, xbuf);
 }
 
 static void
@@ -411,7 +591,7 @@ get_sfp_rx_power(struct i2c_info *ii, ch
 
 	memset(xbuf, 0, sizeof(xbuf));
 	ii->f(ii, SFF_8472_DIAG, SFF_8472_RX_POWER, 2, xbuf);
-	convert_power(ii, xbuf, buf, size);
+	convert_sff_power(ii, buf, size, xbuf);
 }
 
 static void
@@ -421,7 +601,27 @@ get_sfp_tx_power(struct i2c_info *ii, ch
 
 	memset(xbuf, 0, sizeof(xbuf));
 	ii->f(ii, SFF_8472_DIAG, SFF_8472_TX_POWER, 2, xbuf);
-	convert_power(ii, xbuf, buf, size);
+	convert_sff_power(ii, buf, size, xbuf);
+}
+
+static void
+get_qsfp_rx_power(struct i2c_info *ii, char *buf, size_t size, int chan)
+{
+	char xbuf[2];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_RX_CH1_MSB + (chan - 1) * 2, 2, xbuf);
+	convert_sff_power(ii, buf, size, xbuf);
+}
+
+static void
+get_qsfp_tx_power(struct i2c_info *ii, char *buf, size_t size, int chan)
+{
+	char xbuf[2];
+
+	memset(xbuf, 0, sizeof(xbuf));
+	ii->f(ii, SFF_8436_BASE, SFF_8436_TX_CH1_MSB + (chan -1) * 2, 2, xbuf);
+	convert_sff_power(ii, buf, size, xbuf);
 }
 
 /* Intel ixgbe-specific structures and handlers */
@@ -463,50 +663,127 @@ read_i2c_ixgbe(struct i2c_info *ii, uint
 	return (0);
 }
 
-void
-sfp_status(int s, struct ifreq *ifr, int verbose)
+/* Generic handler */
+static int
+read_i2c_generic(struct i2c_info *ii, uint8_t addr, uint8_t off, uint8_t len,
+    caddr_t buf)
+{
+
+	ii->error = EINVAL;
+	return (-1);
+}
+
+static void
+print_qsfp_status(struct i2c_info *ii, int verbose)
 {
-	struct i2c_info ii;
 	char buf[80], buf2[40], buf3[40];
+	uint8_t diag_type;
+	int i;
+
+	/* Read diagnostic monitoring type */
+	ii->f(ii, SFF_8436_BASE, SFF_8436_DIAG_TYPE, 1, (caddr_t)&diag_type);
+	if (ii->error != 0)
+		return;
 
 	/*
-	 * Check if we have i2c support for particular driver.
-	 * TODO: Determine driver by original name.
+	 * Read monitoring data it is supplied.
+	 * XXX: It is not exactly clear from standard
+	 * how one can specify lack of measurements (passive cables case).
 	 */
-	memset(&ii, 0, sizeof(ii));
-	if (strncmp(ifr->ifr_name, "ix", 2) == 0) {
-		ii.f = read_i2c_ixgbe;
-	} else
-		return;
+	if (diag_type != 0)
+		ii->do_diag = 1;
+	ii->qsfp = 1;
 
-	/* Prepare necessary into to pass to NIC handler */
-	ii.s = s;
-	ii.ifr = ifr;
+	/* Transceiver type */
+	get_qsfp_identifier(ii, buf, sizeof(buf));
+	get_qsfp_transceiver_class(ii, buf2, sizeof(buf2));
+	get_qsfp_connector(ii, buf3, sizeof(buf3));
+	if (ii->error == 0)
+		printf("\tplugged: %s %s (%s)\n", buf, buf2, buf3);
+	print_sfp_vendor(ii, buf, sizeof(buf));
+	if (ii->error == 0)
+		printf("\t%s\n", buf);
+
+	/* Request current measurements if they are provided: */
+	if (ii->do_diag != 0) {
+		get_qsfp_temp(ii, buf, sizeof(buf));
+		get_qsfp_voltage(ii, buf2, sizeof(buf2));
+		printf("\tmodule temperature: %s voltage: %s\n", buf, buf2);
+		for (i = 1; i <= 4; i++) {
+			get_qsfp_rx_power(ii, buf, sizeof(buf), i);
+			get_qsfp_tx_power(ii, buf2, sizeof(buf2), i);
+			printf("\tlane %d: RX: %s TX: %s\n", i, buf, buf2);
+		}
+	}
+}
+
+static void
+print_sfp_status(struct i2c_info *ii, int verbose)
+{
+	char buf[80], buf2[40], buf3[40];
+	uint8_t diag_type, flags;
 
 	/* Read diagnostic monitoring type */
-	ii.f(&ii, SFF_8472_BASE, SFF_8472_DIAG_TYPE, 1, (caddr_t)&ii.diag_type);
+	ii->f(ii, SFF_8472_BASE, SFF_8472_DIAG_TYPE, 1, (caddr_t)&diag_type);
+	if (ii->error != 0)
+		return;
+
+	/*
+	 * Read monitoring data IFF it is supplied AND is
+	 * internally calibrated
+	 */
+	flags = SFF_8472_DDM_DONE | SFF_8472_DDM_INTERNAL;
+	if ((diag_type & flags) == flags)
+		ii->do_diag = 1;
 
 	/* Transceiver type */
-	get_sfp_identifier(&ii, buf, sizeof(buf));
-	get_sfp_transceiver_class(&ii, buf2, sizeof(buf2));
-	get_sfp_connector(&ii, buf3, sizeof(buf3));
-	if (ii.error == 0)
-		printf("\ti2c: %s %s (%s)\n", buf, buf2, buf3);
+	get_sfp_identifier(ii, buf, sizeof(buf));
+	get_sfp_transceiver_class(ii, buf2, sizeof(buf2));
+	get_sfp_connector(ii, buf3, sizeof(buf3));
+	if (ii->error == 0)
+		printf("\tplugged: %s %s (%s)\n", buf, buf2, buf3);
 	if (verbose > 2)
-		printf_sfp_transceiver_descr(&ii, buf, sizeof(buf));
-	print_sfp_vendor(&ii, buf, sizeof(buf));
-	if (ii.error == 0)
+		printf_sfp_transceiver_descr(ii, buf, sizeof(buf));
+	print_sfp_vendor(ii, buf, sizeof(buf));
+	if (ii->error == 0)
 		printf("\t%s\n", buf);
 
 	/*
 	 * Request current measurements iff they are provided:
-	 * Bit 6 must be set.
 	 */
-	if ((ii.diag_type & 0x40) != 0) {
-		get_sfp_temp(&ii, buf, sizeof(buf));
-		get_sfp_rx_power(&ii, buf2, sizeof(buf2));
-		get_sfp_tx_power(&ii, buf3, sizeof(buf3));
-		printf("\tTemp: %s RX: %s TX: %s\n", buf, buf2, buf3);
+	if (ii->do_diag != 0) {
+		get_sfp_temp(ii, buf, sizeof(buf));
+		get_sfp_voltage(ii, buf2, sizeof(buf2));
+		printf("\tmodule temperature: %s Voltage: %s\n", buf, buf2);
+		get_sfp_rx_power(ii, buf, sizeof(buf));
+		get_sfp_tx_power(ii, buf2, sizeof(buf2));
+		printf("\tRX: %s TX: %s\n", buf, buf2);
 	}
 }
 
+void
+sfp_status(int s, struct ifreq *ifr, int verbose)
+{
+	struct i2c_info ii;
+
+	/* Prepare necessary into to pass to NIC handler */
+	ii.s = s;
+	ii.ifr = ifr;
+
+	/*
+	 * Check if we have i2c support for particular driver.
+	 * TODO: Determine driver by original name.
+	 */
+	memset(&ii, 0, sizeof(ii));
+	if (strncmp(ifr->ifr_name, "ix", 2) == 0) {
+		ii.f = read_i2c_ixgbe;
+		print_sfp_status(&ii, verbose);
+	} else if (strncmp(ifr->ifr_name, "cxl", 3) == 0) {
+		ii.port_id = atoi(&ifr->ifr_name[3]);
+		ii.f = read_i2c_generic;
+		ii.cfd = -1;
+		print_qsfp_status(&ii, verbose);
+	} else
+		return;
+}
+

Added: head/sys/net/sff8436.h
==============================================================================
--- /dev/null	00:00:00 1970	(empty, because file is newly added)
+++ head/sys/net/sff8436.h	Thu Aug 21 17:54:42 2014	(r270287)
@@ -0,0 +1,211 @@
+/*-
+ * Copyright (c) 2014 Yandex LLC.
+ *
+ * All rights reserved.
+ *
+ * Redistribution and use in source and binary forms, with or without
+ * modification, are permitted provided that the following conditions
+ * are met:
+ * 1. Redistributions of source code must retain the above copyright
+ *    notice, this list of conditions and the following disclaimer.
+ * 2. Redistributions in binary form must reproduce the above copyright
+ *    notice, this list of conditions and the following disclaimer in the
+ *    documentation and/or other materials provided with the distribution.
+ *
+ * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
+ * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+ * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
+ * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
+ * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+ * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
+ * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
+ * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
+ * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
+ * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
+ * SUCH DAMAGE.
+ *
+ * $FreeBSD$
+ */
+
+/*
+ * The following set of constants are from Document SFF-8436
+ * "QSFP+ 10 Gbs 4X PLUGGABLE TRANSCEIVER" revision 4.8 dated October 31, 2013
+ *
+ * This SFF standard defines the following QSFP+ memory address module:
+ *
+ * 1) 256-byte addressable block and 128-byte pages
+ * 2) Lower 128-bytes addresses always refer to the same page
+ * 3) Upper address space may refer to different pages depending on
+ *   "page select" byte value.
+ *
+ * Map description:
+ *
+ * Serial address 0xA02:
+ *
+ * Lower bits
+ * 0-127   Monitoring data & page select byte
+ * 128-255:
+ *
+ * Page 00:
+ * 128-191 Base ID Fields
+ * 191-223 Extended ID
+ * 223-255 Vendor Specific ID
+ *
+ * Page 01 (optional):
+ * 128-255 App-specific data
+ *
+ * Page 02 (optional):
+ * 128-255 User EEPROM Data
+ *
+ * Page 03 (optional for Cable Assmeblies)
+ * 128-223 Thresholds
+ * 225-237 Vendor Specific
+ * 238-253 Channel Controls/Monitor
+ * 254-255 Reserverd
+ *
+ * All these values are read across an I2C (i squared C) bus.
+ */
+
+#define	SFF_8436_BASE	0xA0	/* Base address for all requests */
+
+/* Table 17 - Lower Memory Map */
+enum {
+	SFF_8436_MID		= 0,	/* Copy of SFF_8436_ID field */
+	SFF_8436_STATUS		= 1,	/* 2-bytes status (Table 18) */
+	SFF_8436_INTR_START	= 3,	/* Interrupt flags (Tables 19-21) */
+	SFF_8436_INTR_END	= 21,
+	SFF_8436_MODMON_START	= 22,	/* Module monitors (Table 22 */
+	SFF_8436_TEMP		= 22,	/* Internally measured module temp */
+	SFF_8436_VCC		= 26,	/* Internally mesasure module
+					* supplied voltage */
+	SFF_8436_MODMON_END	= 33,
+	SFF_8436_CHMON_START	= 34,	/* Channel monitors (Table 23) */
+	SFF_8436_RX_CH1_MSB	= 34,	/* Internally measured RX input power */
+	SFF_8436_RX_CH1_LSB	= 35,	/* for channel 1 */
+	SFF_8436_RX_CH2_MSB	= 36,	/* Internally measured RX input power */
+	SFF_8436_RX_CH2_LSB	= 37,	/* for channel 2 */
+	SFF_8436_RX_CH3_MSB	= 38,	/* Internally measured RX input power */
+	SFF_8436_RX_CH3_LSB	= 39,	/* for channel 3 */
+	SFF_8436_RX_CH4_MSB	= 40,	/* Internally measured RX input power */
+	SFF_8436_RX_CH4_LSB	= 41,	/* for channel 4 */
+	SFF_8436_TX_CH1_MSB	= 42,	/* Internally measured TX bias */
+	SFF_8436_TX_CH1_LSB	= 43,	/* for channel 1 */
+	SFF_8436_TX_CH2_MSB	= 44,	/* Internally measured TX bias */
+	SFF_8436_TX_CH2_LSB	= 45,	/* for channel 2 */
+	SFF_8436_TX_CH3_MSB	= 46,	/* Internally measured TX bias */
+	SFF_8436_TX_CH3_LSB	= 47,	/* for channel 3 */
+	SFF_8436_TX_CH4_MSB	= 48,	/* Internally measured TX bias */
+	SFF_8436_TX_CH4_LSB	= 49,	/* for channel 4 */
+	SFF_8436_CHANMON_END	= 81,
+	SFF_8436_CONTROL_START	= 86,	/* Control (Table 24) */
+	SFF_8436_CONTROL_END	= 97,
+	SFF_8436_MASKS_START	= 100,	/* Module/channel masks (Table 25) */
+	SFF_8436_MASKS_END	= 106,
+	SFF_8436_CHPASSWORD	= 119,	/* Password change entry (4 bytes) */
+	SFF_8436_PASSWORD	= 123,	/* Password entry area (4 bytes) */
+	SFF_8436_PAGESEL	= 127,	/* Page select byte */
+};
+
+/* Table 18 - Status Indicators bits */
+/* Byte 1: all bits reserved */
+
+/* Byte 2 bits */
+#define	SFF_8436_STATUS_FLATMEM	(1 << 2)	/* Upper memory flat or paged
+						* 0 = paging, 1=Page 0 only */
+#define	SFF_8436_STATUS_INTL	(1 << 1)	/* Digital state of the intL
+						* Interrupt output pin */
+#define	SFF_8436_STATUS_NOTREADY 1		/* Module has not yet achieved
+						* power up and memory data is not
+						* ready. 0=data is ready */
+/*
+ * Upper page 0 definitions:
+ * Table 29 - Serial ID: Data fields.
+ *
+ * Note that this table is mostly the same as used in SFF-8472.
+ * The only differenee is address shift: +128 bytes.
+ */
+enum {
+	SFF_8436_ID		= 128,  /* Module Type (defined in sff8472.h) */
+	SFF_8436_EXT_ID		= 129,  /* Extended transceiver type
+					 * (Table 31) */
+	SFF_8436_CONNECTOR	= 130,  /* Connector type (Table 32) */
+	SFF_8436_TRANS_START	= 131,  /* Electric or Optical Compatibility
+					 * (Table 33) */
+	SFF_8436_CODE_E1040G	= 131,	/* 10/40G Ethernet Compliance Code */
+	SFF_8436_CODE_SONET	= 132,	/* SONET Compliance codes */
+	SFF_8436_CODE_SATA	= 133,	/* SAS/SATA compliance codes */
+	SFF_8436_CODE_E1G	= 134,	/* Gigabit Ethernet Compliant codes */
+	SFF_8436_CODE_FC_START	= 135,	/* FC link/media/speed */
+	SFF_8436_CODE_FC_END	= 138,
+	SFF_8436_TRANS_END	= 138,
+	SFF_8436_ENCODING	= 139,	/* Encoding Code for high speed
+					* serial encoding algorithm (see
+					* Table 34) */
+	SFF_8436_BITRATE	= 140,	/* Nominal signaling rate, units
+					* of 100MBd. */
+	SFF_8436_RATEID		= 141,	/* Extended RateSelect Compliance
+					* (see Table 35) */
+	SFF_8436_LEN_SMF_KM	= 142,	/* Link length supported for single
+					* mode fiber, units of km */
+	SFF_8436_LEN_OM3	= 143,	/* Link length supported for 850nm
+					* 50um multimode fiber, units of 2 m */
+	SFF_8436_LEN_OM2	= 144, 	/* Link length supported for 50 um
+					* OM2 fiber, units of 1 m */
+	SFF_8436_LEN_OM1	= 145,	/* Link length supported for 1310 nm
+					 * 50um multi-mode fiber, units of 1m*/
+	SFF_8436_LEN_ASM	= 144, /* Link length of passive cable assembly
+					* Length is specified as in the INF
+					* 8074, units of 1m. 0 means this is
+					* not value assembly. Value of 255
+					* means thet the Module supports length
+					* greater than 254 m. */
+	SFF_8436_DEV_TECH	= 147,	/* Device/transmitter technology,
+					* see Table 36/37 */
+	SFF_8436_VENDOR_START	= 148,	/* Vendor name, 16 bytes, padded
+					* right with 0x20 */
+	SFF_8436_VENDOR_END	= 163,
+	SFF_8436_EXTMODCODE	= 164,	/* Extended module code, Table 164 */
+	SFF_8436_VENDOR_OUI_START	= 165 , /* Vendor OUI SFP vendor IEEE
+					* company ID */
+	SFF_8436_VENDOR_OUI_END	= 167,
+	SFF_8436_PN_START 	= 168,	/* Vendor PN, padded right with 0x20 */
+	SFF_8436_PN_END 	= 183,
+	SFF_8436_REV_START 	= 184,	/* Vendor Revision, padded right 0x20 */
+	SFF_8436_REV_END 	= 185,
+	SFF_8436_WAVELEN_START	= 186,	/* Wavelength Laser wavelength
+					* (Passive/Active Cable
+					* Specification Compliance) */
+	SFF_8436_WAVELEN_END	= 189,
+	SFF_8436_MAX_CASE_TEMP	= 190,	/* Allows to specify maximum temp
+					* above 70C. Maximum case temperature is
+					* an 8-bit value in Degrees C. A value
+					*of 0 implies the standard 70C rating.*/
+	SFF_8436_CC_BASE	= 191,	/* CC_BASE Check code for Base ID
+					* Fields (first 63 bytes) */
+	/* Extended ID fields */
+	SFF_8436_OPTIONS_START	= 192, /* Options Indicates which optional
+					* transceiver signals are
+					* implemented (see Table 39) */
+	SFF_8436_OPTIONS_END	= 195,
+	SFF_8436_SN_START 	= 196,	/* Vendor SN, riwght padded with 0x20 */
+	SFF_8436_SN_END 	= 211,
+	SFF_8436_DATE_START	= 212,	/* Vendor’s manufacturing date code
+					* (see Table 40) */
+	SFF_8436_DATE_END	= 219,
+	SFF_8436_DIAG_TYPE	= 220,	/* Diagnostic Monitoring Type
+					* Indicates which type of
+					* diagnostic monitoring is
+					* implemented (if any) in the
+					* transceiver (see Table 41) */
+
+	SFF_8436_ENHANCED	= 221,	/* Enhanced Options Indicates which
+					* optional features are implemented
+					* (if any) in the transceiver
+					* (see Table 42) */
+	SFF_8436_CC_EXT		= 222,	/* Check code for the Extended ID
+					* Fields (bytes 192-222 incl) */
+	SFF_8436_VENDOR_RSRVD_START	= 224,
+	SFF_8436_VENDOR_RSRVD_END	= 255,
+};
+
+

Modified: head/sys/net/sff8472.h
==============================================================================
--- head/sys/net/sff8472.h	Thu Aug 21 17:36:42 2014	(r270286)
+++ head/sys/net/sff8472.h	Thu Aug 21 17:54:42 2014	(r270287)
@@ -375,37 +375,98 @@ enum {
  */
 #define SFF_8472_STATUS_DATA_READY (1 << 0)
 
-/* Table 3.2 Identifier values */
+/*
+ * Table 3.2 Identifier values.
+ * Identifier constants has taken from SFF-8024 rev 2.2 table 4.1
+ * (as referenced by table 3.2 footer)
+ * */
 enum {
-	SFF_8472_ID_UNKNOWN	= 0x0, /* Unknown or unspecified */
-	SFF_8472_ID_GBIC	= 0x1, /* GBIC */
-	SFF_8472_ID_SFF		= 0x2, /* Module soldered to motherboard (ex: SFF)*/
-	SFF_8472_ID_SFP		= 0x3, /* SFP or SFP “Plus” */
-	SFF_8472_ID_XBI		= 0x4, /* Reserved for “300 pin XBI” devices */
-	SFF_8472_ID_XENPAK	= 0x5, /* Reserved for “Xenpak” devices */
-	SFF_8472_ID_XFP		= 0x6, /* Reserved for “XFP” devices */
-	SFF_8472_ID_XFF		= 0x7, /* Reserved for “XFF” devices */
-	SFF_8472_ID_XFPE	= 0x8, /* Reserved for “XFP-E” devices */
-	SFF_8472_ID_XPAK	= 0x9, /* Reserved for “XPak” devices */
-	SFF_8472_ID_X2		= 0xA, /* Reserved for “X2” devices */
-	SFF_8472_ID_DWDM_SFP	= 0xB, /* Reserved for “DWDM-SFP” devices */
-	SFF_8472_ID_QSFP	= 0xC, /* Reserved for “QSFP” devices */
-	SFF_8472_ID_LAST	= SFF_8472_ID_QSFP
+	SFF_8024_ID_UNKNOWN	= 0x0, /* Unknown or unspecified */
+	SFF_8024_ID_GBIC	= 0x1, /* GBIC */
+	SFF_8024_ID_SFF		= 0x2, /* Module soldered to motherboard (ex: SFF)*/
+	SFF_8024_ID_SFP		= 0x3, /* SFP or SFP “Plus” */
+	SFF_8024_ID_XBI		= 0x4, /* 300 pin XBI */
+	SFF_8024_ID_XENPAK	= 0x5, /* Xenpak */
+	SFF_8024_ID_XFP		= 0x6, /* XFP */
+	SFF_8024_ID_XFF		= 0x7, /* XFF */
+	SFF_8024_ID_XFPE	= 0x8, /* XFP-E */
+	SFF_8024_ID_XPAK	= 0x9, /* XPAk */
+	SFF_8024_ID_X2		= 0xA, /* X2 */
+	SFF_8024_ID_DWDM_SFP	= 0xB, /* DWDM-SFP */
+	SFF_8024_ID_QSFP	= 0xC, /* QSFP */
+	SFF_8024_ID_QSFPPLUS	= 0xD, /* QSFP+ */
+	SFF_8024_ID_CXP		= 0xE, /* CXP */
+	SFF_8024_ID_HD4X	= 0xF, /* Shielded Mini Multilane HD 4X */ 
+	SFF_8024_ID_HD8X	= 0x10, /* Shielded Mini Multilane HD 8X */ 
+	SFF_8024_ID_QSFP28	= 0x11, /* QSFP28 */
+	SFF_8024_ID_CXP2	= 0x12, /* CXP2 (aka CXP28) */
+	SFF_8024_ID_LAST	= SFF_8024_ID_CXP2
 	};
 
-static const char *sff_8472_id[SFF_8472_ID_LAST + 1] = {"Unknown",
+static const char *sff_8024_id[SFF_8024_ID_LAST + 1] = {"Unknown",
 					     "GBIC",
 					     "SFF",
-					     "SFP",
+					     "SFP/SFP+",
 					     "XBI",
 					     "Xenpak",
 					     "XFP",
 					     "XFF",
 					     "XFP-E",
-					     "XPak",
+					     "XPAk",
 					     "X2",
 					     "DWDM-SFP",
-					     "QSFP"};
+					     "QSFP",
+					     "QSFP+",
+					     "CXP",
+					     "HD4X",
+					     "HD8X",
+					     "QSFP28",
+					     "CXP2"};
+
+/* Keep compability with old definitions */
+#define	SFF_8472_ID_UNKNOWN	SFF_8024_ID_UNKNOWN
+#define	SFF_8472_ID_GBIC	SFF_8024_ID_GBIC
+#define	SFF_8472_ID_SFF		SFF_8024_ID_SFF
+#define	SFF_8472_ID_SFP		SFF_8024_ID_SFP
+#define	SFF_8472_ID_XBI		SFF_8024_ID_XBI
+#define	SFF_8472_ID_XENPAK	SFF_8024_ID_XENPAK
+#define	SFF_8472_ID_XFP		SFF_8024_ID_XFP
+#define	SFF_8472_ID_XFF		SFF_8024_ID_XFF
+#define	SFF_8472_ID_XFPE	SFF_8024_ID_XFPE
+#define	SFF_8472_ID_XPAK	SFF_8024_ID_XPAK
+#define	SFF_8472_ID_X2		SFF_8024_ID_X2
+#define	SFF_8472_ID_DWDM_SFP	SFF_8024_ID_DWDM_SFP
+#define	SFF_8472_ID_QSFP	SFF_8024_ID_QSFP
+#define	SFF_8472_ID_LAST	SFF_8024_ID_LAST
+
+#define	sff_8472_id		sff_8024_id
+
+/*
+ * Table 3.9 Diagnostic Monitoring Type (byte 92)
+ * bits described.
+ */
+
+/*
+ * Digital diagnostic monitoring implemented.
+ * Set to 1 for transceivers implementing DDM.
+ */
+#define	SFF_8472_DDM_DONE	(1 << 6)
+
+/*
+ * Measurements are internally calibrated.
+ */
+#define	SFF_8472_DDM_INTERNAL	(1 << 5)
+
+/*
+ * Measurements are externally calibrated.
+ */
+#define	SFF_8472_DDM_EXTERNAL	(1 << 4)
+
+/*
+ * Received power measurement type
+ * 0 = OMA, 1 = average power
+ */
+#define	SFF_8472_DDM_PMTYPE	(1 << 3)
 
 /* Table 3.13 and 3.14 Temperature Conversion Values */
 #define SFF_8472_TEMP_SIGN (1 << 15)


More information about the svn-src-all mailing list